/*
        Copyright (c) 2004 on behalf of the EU EGEE Project:
        The European Organization for Nuclear Research (CERN),
        Istituto Nazionale di Fisica Nucleare (INFN), Italy
        Datamat Spa, Italy
        Centre National de la Recherche Scientifique (CNRS), France
        CS Systeme d'Information (CSSI), France
        Royal Institute of Technology, Center for Parallel Computers (KTH-PDC), Sweden
        Universiteit van Amsterdam (UvA), Netherlands
        University of Helsinki (UH.HIP), Finland
        University of Bergen (UiB), Norway
        Council for the Central Laboratory of the Research Councils (CCLRC), United Kingdom

        JobInfoWrapper: a wrapper for the new class Info
        
        Authors: Alvise Dorigo <alvise.dorigo@pd.infn.it>
        Version info: $Id: VOMSWrapper.cpp,v 1.7.2.3 2008/07/28 15:47:14 adorigo Exp $
*/

#include "glite/ce/cream-client-api-c/VOMSWrapper.h"
#include <vector>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <exception>
#include <iostream>
#include <openssl/pem.h>
#include <cerrno>
#include "glite/ce/cream-client-api-c/creamApiLogger.h"

#include <boost/shared_ptr.hpp>

#include <time.h>

extern int daylight;
extern long timezone;

using namespace glite::ce::cream_client_api::soap_proxy;
using namespace std;

extern int *__errno_location( void );
#define errno (*__errno_location())

boost::recursive_mutex VOMSWrapper::s_mutex;

//______________________________________________________________________________
VOMSWrapper::VOMSWrapper( const std::string& pxfile ) 
  throw() : m_pxfile ( pxfile )
{
  boost::recursive_mutex::scoped_lock M( s_mutex );

  m_isValid = false;
  
  FILE* certIN;
  int saveerr;
  certIN = fopen( pxfile.c_str(), "r" );
  
  if( !certIN ) {
    saveerr = errno;
    m_errorMessage = "VOMSWrapper::CTOR() - Unable to open the file [";
    m_errorMessage += pxfile + "] : " + strerror( saveerr );
    m_errorNum = FILE_NOT_FOUND;
    return;
  }
  
  boost::shared_ptr<std::FILE> tmpIn(certIN, std::fclose);
  
  try {
    this->timeEnd();
  }catch(exception& ex) {
    m_errorMessage = ex.what();
    m_errorNum = VALIDITY_ERROR;
    return;
  }

  if( this->getProxyTimeEnd() <= time(NULL) ) {
    m_errorNum = EXPIRED_PROXY;
    m_errorMessage = "VOMSWrapper::CTOR() - The proxy is EXPIRED!";
    //CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->errorStream()
     //            << "VOMSWrapper::CTOR() - Proxy is EXPIRED!"
      //           << log4cpp::eol);

    return;
  }

  // do set yet m_isValid to true because
  // we must before check the validity
  // of internal attribute certificate

  
  // VOMS API do not accept an expired proxy
  
  int res = m_V.Retrieve( certIN, RECURSE_CHAIN );
  
  if( !res ) {
    
    m_V.SetVerificationType( (verify_type)(VERIFY_NONE) );
    res = m_V.Retrieve( certIN, RECURSE_CHAIN );
    
    if(!res) {
      m_errorMessage = "Unable to verify attribute certificate signature! Error is: [";
      m_errorMessage += m_V.ErrorMessage() + "]";
      
      m_errorNum = VALIDITY_ERROR;
      m_isValid = false;
    } else
      m_isValid = true;
    
  } else {
    m_isValid = true;
  }
  
  //fclose( certIN );
}

//______________________________________________________________________________
string VOMSWrapper::getVOName( void ) const 
{ 
  if( m_V.data.begin() != m_V.data.end() )
    return m_V.data.begin()->voname; 
  else return "";
}

//______________________________________________________________________________
string VOMSWrapper::getIssuer( void ) const
{
  if( m_V.data.begin() != m_V.data.end() )
    return m_V.data.begin()->user; 
  else return "";
}

//______________________________________________________________________________
std::string VOMSWrapper::getDNFQAN( void ) const throw () {
  
  std::vector<std::string> fqans;
  std::string fqan = "";
  
  this->getFQANs( fqans );
  
  for(std::vector<std::string>::const_iterator it=fqans.begin(); it!=fqans.end(); ++it)
    {
      fqan += *it;    
    }
  
  return this->getSubject( ) 
    + "-" 
    + fqan; // both methods can raise an auth_ex exception
  
}

//______________________________________________________________________________
void VOMSWrapper::timeEnd( void ) throw(auth_ex&) 
{
  
  X509 *x;
  x = VOMSWrapper::read_BIO( ); // can raise a auth_ex
  VOMSWrapper::helper_X509new X(x);
  ASN1_TIME *ctm = X509_get_notAfter(x);
  	
 

  int len = ctm->length;
  char *buf = (char*)malloc(len+1);
  memset((void*)buf, 0, len+1);
  memcpy(buf, ctm->data, len);
//   CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->debugStream()
//                  << "VOMSWrapper::timeEnd() - X509_get_notAfter RETURNED ["
// 	         << buf << "]"
//                  << log4cpp::eol);
		 
  // ::setenv("TZ", "", 1);
  struct tm target;
  strptime( buf, "%y%m%d%H%M%SZ", &target );
  
//   CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->debugStream()
//                  << "VOMSWrapper::timeEnd() - PARSED ["
// 	         << buf << "] to time_t [" << m_proxyTimeEnd << "]"
//                  << log4cpp::eol);
  
  free( buf );
  m_proxyTimeEnd = mktime( &target );

  if(daylight!=0)
    m_proxyTimeEnd += 3600;

  m_proxyTimeEnd -= timezone;

  //   cout <<"proxyLeft = "<<(m_proxyTimeEnd-time(0))<<endl;
//  cout << "daylight="<<daylight<< " - timezone="<<timezone<<endl;
//   CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->debugStream()
//                  << "VOMSWrapper::timeEnd() - Time_t NOW is ["
// 	         << time(NULL) << "]"
//                  << log4cpp::eol);	
		 
//  ::unsetenv("TZ");

}

//______________________________________________________________________________
X509* VOMSWrapper::read_BIO( void ) 
  throw(auth_ex&)
{
  BIO *in = NULL;
  in = BIO_new(BIO_s_file());
  VOMSWrapper::helper_BIOnew h( in ); // smart container to 
                                      // free 'in' when leaving this scope
  if (in) {
    if (BIO_read_filename( in, m_pxfile.c_str() ) > 0) {
        
      X509* x = PEM_read_bio_X509(in, NULL, 0, NULL);
      if(!x) {
        throw  auth_ex(std::string("VOMSWrapper::read_BIO() - PEM_read_bio_X509 failed while reading X509 proxy file: ")+ m_pxfile)
;
      }
      
      return x;

    } else {
      throw  auth_ex(std::string("VOMSWrapper::read_BIO() - BIO_read_filename failed while reading X509 proxy file: ")+ m_pxfile);
    }
  } else {
    throw  auth_ex(std::string("VOMSWrapper::read_BIO() - unable to allocate memory when creatin a BIO object with BIO_new() call, for the proxy file: ")+ m_pxfile);
  }
}
