/*
        Copyright (c) 2004 on behalf of the EU EGEE Project:
        The European Organization for Nuclear Research (CERN),
        Istituto Nazionale di Fisica Nucleare (INFN), Italy
        Datamat Spa, Italy
        Centre National de la Recherche Scientifique (CNRS), France
        CS Systeme d'Information (CSSI), France
        Royal Institute of Technology, Center for Parallel Computers (KTH-PDC), Sweden
        Universiteit van Amsterdam (UvA), Netherlands
        University of Helsinki (UH.HIP), Finland
        University of Bergen (UiB), Norway
        Council for the Central Laboratory of the Research Councils (CCLRC), United Kingdom

        certUtil: utilities to extract useful info from a proxy certificate
	
        Authors: Alvise Dorigo <alvise.dorigo@pd.infn.it>
        Version info: $Id: certUtil.cpp,v 1.19.2.1 2008/06/10 07:21:55 adorigo Exp $
*/

#include "glite/ce/cream-client-api-c/certUtil.h"
#include "glite/ce/cream-client-api-c/creamApiLogger.h"

extern "C" {
  #include "gridsite.h"
}

#include "boost/algorithm/string.hpp"
#include "boost/format.hpp"

#include <vector>
#include <ctime>
#include <unistd.h>
#include <openssl/sha.h>

namespace soapProxy = glite::ce::cream_client_api::soap_proxy;

void getFQANs(const std::string& pxfile, std::vector<std::string>& vect) throw (soapProxy::auth_ex&);
			   
time_t ASN1_UTCTIME_get(const ASN1_UTCTIME *s);

#define GRST_MAX_CHAIN_LEN 9

static void mpcerror(FILE *debugfp, char *msg)
{
  if (debugfp != NULL)
    {
      fputs(msg, debugfp);
      //ERR_print_errors_fp(debugfp);
      CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->errorStream()
		     << msg
		     );

      fprintf( debugfp, "%s", msg);

    }
}

// namespace {
//   
//   class helper_BIOnew {
//     BIO *in;
//   public:
//     helper_BIOnew(BIO* infile) : in(infile) {}
//     ~helper_BIOnew() { 
//       if(in) {
//         BIO_free(in);
//       } 
//     }
//   };
// 
//   class helper_X509new {
//     X509* x;
//   public:
//     helper_X509new(X509* xin) : x(xin) {}
//     ~helper_X509new() { if(x) X509_free(x); }
//   };
// 
//   class helper_X509_NAME_new {
//     X509_NAME* x;
//   public:
//     helper_X509_NAME_new(X509_NAME* xin) : x(xin) {}
//     ~helper_X509_NAME_new() { if(x) X509_NAME_free(x); }
//   };
// 
//   class helper_X509_REQ_new {
//     X509_REQ* x;
//   public:
//     helper_X509_REQ_new(X509_REQ* xin) : x(xin) {}
//     ~helper_X509_REQ_new() { if(x) X509_REQ_free(x); }
//   };
// 
// }

#ifndef GRST_BACKDATE_SECONDS
#define GRST_BACKDATE_SECONDS 300
//#define GRST_BACKDATE_SECONDS 0
#endif

/// Make a GSI Proxy chain from a request, certificate and private key
int glite::ce::cream_client_api::certUtil::GRSTx509MakeProxyCert_local(char **proxychain, FILE *debugfp, 
                          char *reqtxt, char *cert, char *key, int minutes)
///
/// The proxy chain is returned in *proxychain. If debugfp is non-NULL,
/// errors are output to that file pointer. The proxy will expired in
/// the given number of minutes starting from the current time.
{
  char *ptr, *certchain;
  int i, ncerts;
  long serial = 1234, ptrlen;
  EVP_PKEY *pkey, *CApkey;
  const EVP_MD *digest;
  X509 *certs[GRST_MAX_CHAIN_LEN];
  X509_REQ *req;
  X509_NAME *name, *CAsubject, *newsubject;
  X509_NAME_ENTRY *ent;
  FILE *fp;
  BIO *reqmem, *certmem;
  time_t notAfter;
  time_t notBefore;

  /* read in the request */
  reqmem = BIO_new(BIO_s_mem());
  BIO_puts(reqmem, reqtxt);
    
  if (!(req = PEM_read_bio_X509_REQ(reqmem, NULL, NULL, NULL)))
    {
      mpcerror(debugfp,
              "GRSTx509MakeProxyCert(): error reading request from BIO memory\n");
      BIO_free(reqmem);
      return GRST_RET_FAILED;
    }
    
  BIO_free(reqmem);

  /* verify signature on the request */
  if (!(pkey = X509_REQ_get_pubkey(req)))
    {
      mpcerror(debugfp,
              "GRSTx509MakeProxyCert(): error getting public key from request\n");
      
      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }

  if (X509_REQ_verify(req, pkey) != 1)
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error verifying signature on certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }
    
  /* read in the signing certificate */
  if (!(fp = fopen(cert, "r")))
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error opening signing certificate file\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  for (ncerts = 1; ncerts < GRST_MAX_CHAIN_LEN; ++ncerts)
   if ((certs[ncerts] = PEM_read_X509(fp, NULL, NULL, NULL)) == NULL) break;

  if (ncerts == 1) /* zeroth cert with be new proxy cert */
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error reading signing certificate file\n");

      X509_REQ_free(req);
      fclose( fp );
      return GRST_RET_FAILED;
    }    

  fclose(fp);
  
  CAsubject = X509_get_subject_name(certs[1]);

  /* read in the CA private key */
  if (!(fp = fopen(key, "r")))
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error reading signing private key file\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  if (!(CApkey = PEM_read_PrivateKey(fp, NULL, NULL, NULL)))
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error reading signing private key in file\n");

      X509_REQ_free(req);
      fclose( fp );
      return GRST_RET_FAILED;
    }    

  fclose(fp);
  
  /* get subject name */
  if (!(name = X509_REQ_get_subject_name(req)))
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error getting subject name from request\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  /* create new certificate */
  if (!(certs[0] = X509_new()))
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error creating X509 object\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  /* set version number for the certificate (X509v3) and the serial number   
     need 3 = v4 for GSI proxy?? */
  if (X509_set_version(certs[0], 3L) != 1)
    {
      mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error setting certificate version\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  ASN1_INTEGER_set(X509_get_serialNumber(certs[0]), serial++);

  if (!(name = X509_get_subject_name(certs[1])))
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error getting subject name from CA certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  if (X509_set_issuer_name(certs[0], name) != 1)
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error setting issuer name of certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  /* set issuer and subject name of the cert from the req and the CA */
  ent = X509_NAME_ENTRY_create_by_NID(NULL, OBJ_txt2nid("commonName"), 
                                      MBSTRING_ASC, (unsigned char*)"proxy", -1);

  newsubject = X509_NAME_dup(CAsubject);

  X509_NAME_add_entry(newsubject, ent, -1, 0);

  if (X509_set_subject_name(certs[0], newsubject) != 1)
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error setting subject name of certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    
    
  X509_NAME_free(newsubject);
  X509_NAME_ENTRY_free(ent);

  /* set public key in the certificate */
  if (X509_set_pubkey(certs[0], pkey) != 1)
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error setting public key of the certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  /* set duration for the certificate */
  if (!(X509_gmtime_adj(X509_get_notBefore(certs[0]), -GRST_BACKDATE_SECONDS)))
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error setting beginning time of the certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  if (!(X509_gmtime_adj(X509_get_notAfter(certs[0]), 60 * minutes)))
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error setting ending time of the certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }
    
  /* go through chain selecting the highest start-time*/

  notBefore = 
    GRSTasn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notBefore(certs[0])), 0);

  for (i=1; i < ncerts; ++i)
       if (notBefore < 
           GRSTasn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notBefore(certs[i])),
                               0))
         {
           notBefore = 
            GRSTasn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notBefore(certs[i])),
                                0);
            
           ASN1_UTCTIME_set(X509_get_notBefore(certs[0]), notBefore);
         }

  /* go through chain making sure this proxy is not longer lived */

  notAfter = 
    GRSTasn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notAfter(certs[0])), 0);

  for (i=1; i < ncerts; ++i)
       if (notAfter > 
           GRSTasn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notAfter(certs[i])),
                               0))
         {
           notAfter = 
            GRSTasn1TimeToTimeT((char*)ASN1_STRING_data(X509_get_notAfter(certs[i])),
                                0);
            
           ASN1_UTCTIME_set(X509_get_notAfter(certs[0]), notAfter);
         }

  /* sign the certificate with the signing private key */
  if (EVP_PKEY_type(CApkey->type) == EVP_PKEY_RSA)
    digest = EVP_md5();
  else
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error checking signing private key for a valid digest\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  if (!(X509_sign(certs[0], CApkey, digest)))
    {
      mpcerror(debugfp,
      "GRSTx509MakeProxyCert(): error signing certificate\n");

      X509_REQ_free(req);
      return GRST_RET_FAILED;
    }    

  /* store the completed certificate chain */

  certchain = strdup("");

  for (i=0; i < ncerts; ++i)
     {
       certmem = BIO_new(BIO_s_mem());

       if (PEM_write_bio_X509(certmem, certs[i]) != 1)
         {
           mpcerror(debugfp,
            "GRSTx509MakeProxyCert(): error writing certificate to memory BIO\n");            

           X509_REQ_free(req);
           return GRST_RET_FAILED;
         }

       ptrlen = BIO_get_mem_data(certmem, &ptr);
  
       certchain = (char*)realloc(certchain, strlen(certchain) + ptrlen + 1);
       
       strncat(certchain, ptr, ptrlen);
    
       BIO_free(certmem);
       X509_free(certs[i]);
     }
  
  EVP_PKEY_free(pkey);
  EVP_PKEY_free(CApkey);
  X509_REQ_free(req);
      
  *proxychain = certchain;  
  //proxychain = certchain;
  return GRST_RET_OK;
}

/**
 * END OF PATCH STUFF
 */


//____________________________________________________________________________________
X509* glite::ce::cream_client_api::certUtil::read_BIO( const std::string& pxfile ) 
  throw(soapProxy::auth_ex&)
{
  BIO *in = NULL;
  in = BIO_new(BIO_s_file());
  glite::ce::cream_client_api::helper_BIOnew h( in ); // smart container to 
  						      // free 'in' when leaving this scope
  if (in) {
    if (BIO_read_filename( in, pxfile.c_str() ) > 0) {
	
      X509* x = PEM_read_bio_X509(in, NULL, 0, NULL);
      if(!x) {
	throw  soapProxy::auth_ex(std::string("certUtil::read_BIO() - PEM_read_bio_X509 failed while reading X509 proxy file: ")+ pxfile);
      }
      
      return x;

    } else {
      throw  soapProxy::auth_ex(std::string("certUtil::read_BIO() - BIO_read_filename failed while reading X509 proxy file: ")+ pxfile);
    }
  } else {
    throw  soapProxy::auth_ex(std::string("certUtil::read_BIO() - unable to allocate memory when creatin a BIO object with BIO_new() call, for the proxy file: ")+ pxfile);
  }
}

//____________________________________________________________________________________
std::string glite::ce::cream_client_api::certUtil::getDN(const std::string& pxfile) 
  throw(soapProxy::auth_ex&) 
{
  BIO *in = NULL;
  X509 *x = NULL;
  std::string distinguished_name = "";
  in = BIO_new(BIO_s_file());
  glite::ce::cream_client_api::helper_BIOnew h( in ); // smart container to 
  						      // free 'in' when leaving this scope
  
  if (in) {
    if (BIO_read_filename( in, pxfile.c_str() ) > 0) {
      x = PEM_read_bio_X509(in, NULL, 0, NULL);
      glite::ce::cream_client_api::helper_X509new X(x); // smart container to free 'x'
      							// when leaving this scope    
      if (x) {
	char *name = X509_NAME_oneline(X509_get_subject_name(x), NULL, 0);
	if(name) {
	  distinguished_name = name;
	  free(name);
	} else {
	  throw soapProxy::auth_ex("X509_NAME_oneline returned a NULL pointer!");
	}
      } else{
	throw  soapProxy::auth_ex(std::string("unable to read X509 proxy file: ")+ pxfile);
      }
    } else {
      throw  soapProxy::auth_ex(std::string("unable to open X509 proxy file: ")+ pxfile);
    }
  } else {
    throw  soapProxy::auth_ex(std::string("unable to allocate memory for the proxy file: ")+ pxfile);
  }

  return distinguished_name;
}

//____________________________________________________________________________________
std::string glite::ce::cream_client_api::certUtil::getDNFQAN(const std::string& pxfile) 
  throw(soapProxy::auth_ex&) 
{

  std::vector<std::string> fqans;
  std::string fqan = "";

  glite::ce::cream_client_api::certUtil::getFQANs( pxfile, fqans );

  for(std::vector<std::string>::const_iterator it=fqans.begin(); it!=fqans.end(); ++it)
  {
    fqan += *it;    
  }

  return glite::ce::cream_client_api::certUtil::getCertSubj( pxfile ) 
    + "-" 
    + fqan; // both methods can raise an auth_ex exception

}

//____________________________________________________________________________________
std::string glite::ce::cream_client_api::certUtil::getCertSubj(const std::string& pxfile) 
  throw(soapProxy::auth_ex&) 
{
  std::string Subject;

  X509 *x = glite::ce::cream_client_api::certUtil::read_BIO( pxfile ); // can raise a auth_ex
  glite::ce::cream_client_api::helper_X509new X(x);
  char *name = X509_NAME_oneline(X509_get_subject_name(x), NULL, 0);
  if(name) {
    Subject = name;
    free(name);
  } else {
    throw soapProxy::auth_ex("X509_NAME_oneline returned a NULL pointer!");
  }

  return Subject;
}

//____________________________________________________________________________________
std::string glite::ce::cream_client_api::certUtil::getCertIssuer(const std::string& pxfile) 
  throw(soapProxy::auth_ex&) 
{
  std::string Issuer;

  X509 *x = glite::ce::cream_client_api::certUtil::read_BIO( pxfile ); // can raise a auth_ex
  glite::ce::cream_client_api::helper_X509new X(x);
  char *name = X509_NAME_oneline(X509_get_subject_name(x), NULL, 0);
  if(name) {
    Issuer = name;
    free(name);
  } else {
    throw soapProxy::auth_ex("X509_NAME_oneline returned a NULL pointer!");
  }

  return Issuer;
}

//____________________________________________________________________________________
const long int glite::ce::cream_client_api::certUtil::getProxyTimeLeft( const std::string& pxfile )
  throw (soapProxy::auth_ex&)
{

    long int timeleft = 0;

    X509 *x = glite::ce::cream_client_api::certUtil::read_BIO( pxfile ); // can raise a auth_ex
    glite::ce::cream_client_api::helper_X509new X(x);
    
    timeleft = (long int)((long int)ASN1_UTCTIME_get(X509_get_notAfter(x)) - (long int)time(NULL) );
    return timeleft;
}

//______________________________________________________________________________
X509* glite::ce::cream_client_api::certUtil::getX509Fromfile(const std::string& proxyfile)
  throw(soapProxy::auth_ex&)
{
  BIO* in = NULL;
  in = BIO_new( BIO_s_file() );
  glite::ce::cream_client_api::helper_BIOnew h(in);

  if( !in ) {
    throw soapProxy::auth_ex("certUtil::getX509Fromfile() - BIO_new() returned a NULL pointer");
  }
  X509* x;
  if(BIO_read_filename(in, (char *)proxyfile.c_str()) > 0)
  {
    x = PEM_read_bio_X509( in, NULL, 0, NULL );
    if( !x )
    {
      throw soapProxy::auth_ex( std::string("certUtil::getX509Fromfile() - Cannot read the certificate [")
    			      + proxyfile + "] with PEM_read_bio_X509");
    }
  }
  return x;
}

/*************************************************************************
	private method:	converts ASN1_UTCTIME to time_t
*************************************************************************/
time_t ASN1_UTCTIME_get(const ASN1_UTCTIME *s)
{
  int offset;
  struct tm _tm;
  memset(&_tm,'\0',sizeof(struct tm));
#define g2(p) (((p)[0]-'0')*10+(p)[1]-'0')
  _tm.tm_year=g2(s->data);
  if(_tm.tm_year < 50)
    _tm.tm_year+=100;
  _tm.tm_mon=g2(s->data+2)-1;
  _tm.tm_mday=g2(s->data+4);
  _tm.tm_hour=g2(s->data+6);
  _tm.tm_min=g2(s->data+8);
  _tm.tm_sec=g2(s->data+10);
  if(s->data[12] == 'Z')
    offset=0;
  else
    {
      offset=g2(s->data+13)*60+g2(s->data+15);
      if(s->data[12] == '-')
	offset= -offset;
    }
#undef g2
  
  return timegm(&_tm)-offset*60;
}

//______________________________________________________________________________
STACK_OF(X509) *glite::ce::cream_client_api::certUtil::load_chain(char *certfile) throw()
{
  STACK_OF(X509_INFO) *sk=NULL;
  STACK_OF(X509) *stack=NULL, *ret=NULL;
  BIO *in=NULL;
  X509_INFO *xi;
  int first = 1;

  in=BIO_new_file(certfile, "r");
  glite::ce::cream_client_api::helper_BIOnew h(in);
  stack = sk_X509_new_null();

  if(!in) {
    CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->errorStream() 
		   << "certUtil::load_chain() - error opening the file, "
		   << certfile
		   );
    //printf("load_chain() - error opening the file, %s\n",certfile);
    return NULL;
  }
  
  if(!stack) {
    CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->errorStream() 
		   << "certUtil::load_chain() - memory allocation failure\n"
		   );
    //printf("load_chain() - memory allocation failure\n");
    return NULL;
  }
  
  /* This loads from a file, a stack of x509/crl/pkey sets */
  if(!(sk=PEM_X509_INFO_read_bio(in,NULL,NULL,NULL))) {
    CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->errorStream() 
		   << "certUtil::load_chain() - error reading the file, "
		   << certfile
		   );
    //printf("load_chain() - error reading the file, %s\n",certfile);
    goto end;
  }

  /* scan over it and pull out the certs */
  while (sk_X509_INFO_num(sk)) {
    /* skip first cert */
    if (first) {
      first = 0;
      continue;
    }
    xi=sk_X509_INFO_shift(sk);
    if (xi->x509 != NULL) {
      sk_X509_push(stack,xi->x509);
      xi->x509=NULL;
    }
    X509_INFO_free(xi);
  }
  if(!sk_X509_num(stack)) {
    CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->errorStream() 
		   << "certUtil::load_chain() - no certificates in file, "
		   << certfile
		   );
    //printf("load_chain() - no certificates in file, %s\n",certfile);
    sk_X509_free(stack);
    goto end;
  }
  ret=stack;
end:
  sk_X509_INFO_free(sk);
  return(ret);
}

//_____________________________________________________________________________
void glite::ce::cream_client_api::certUtil::getFQANs(const std::string& filename, std::vector<std::string>& target) 
  throw(soapProxy::auth_ex&)
{

  BIO* in = NULL;
  in = BIO_new( BIO_s_file() );
  glite::ce::cream_client_api::helper_BIOnew h(in);
  
  if( !in ) {
    throw soapProxy::auth_ex("CreamProxy::getX509Fromfile() - BIO_new() returned a NULL pointer");
  }
  X509* x;
  STACK_OF(X509) *sk;
  if(BIO_read_filename(in, (char *)filename.c_str()) > 0)
  {
    x = PEM_read_bio_X509( in, NULL, 0, NULL );
    if(!x) throw soapProxy::auth_ex("PEM_read_bio_X509 failed reading certificate");
    
    glite::ce::cream_client_api::helper_X509new X(x);
    sk = glite::ce::cream_client_api::certUtil::load_chain( (char*)filename.c_str() );
    if(!sk)
      throw soapProxy::auth_ex("load_chain failed reading certificate");
      
    glite::ce::cream_client_api::helper_skX509new skX(sk);
      
    vomsdata vd;
    if (vd.Retrieve(x, sk)) {
     
      target = vd.data.begin()->fqan;

    }
    else {
      throw soapProxy::auth_ex("vomsdata::retrieve returned false!");
    }
  }
}

//___________________________________________________________________________________________
std::string glite::ce::cream_client_api::certUtil::getVOName( const std::string& pxfile )
  throw( soapProxy::auth_ex&)
{

  // FIXME: memory leak, probably related to vomsdata class

  X509 *cert = NULL;
  std::string vo;
  
  cert = glite::ce::cream_client_api::certUtil::getX509Fromfile( pxfile );
  glite::ce::cream_client_api::helper_X509new X(cert);
  
  STACK_OF(X509) *_x = NULL;

  vomsdata V;
  _x = glite::ce::cream_client_api::certUtil::load_chain( (char*)pxfile.c_str() );
  
  glite::ce::cream_client_api::helper_skX509new skX(_x);
  
  if(!_x) return ""; 
  int res = V.Retrieve( cert, _x, RECURSE_CHAIN );
  if( !res ) {
    V.SetVerificationType( (verify_type)(VERIFY_NONE) );
    res = V.Retrieve( cert, _x, RECURSE_CHAIN );
    //std::cerr << "*** WARN: certUtil::getVOName() - Unable to verify certificate signature! Server certificate possibily not installed: "<< V.ErrorMessage()<<std::endl;
    CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->warnStream() 
		   << "certUtil::getVOName() -  Unable to verify certificate signature! "
		   << "Server certificate possibily not installed: "
		   << V.ErrorMessage()
		   );
  }
  
  if( !V.data.empty() ) vo=V.data.begin()->voname;

  return vo;
}

//___________________________________________________________________________________________
void glite::ce::cream_client_api::certUtil::generateUniqueID(std::string& target) throw()
{
  struct timeval Tv;
  struct timezone Tz;
  gettimeofday(&Tv, &Tz);
  char hname[1024];
  char* H;
  if(gethostname(hname, 1024)==-1)
    H = "";
  else
    H = hname;
   
  target = boost::str( boost::format("%d.%d-%d-%d-%s") % Tv.tv_sec % Tv.tv_usec % ::getuid() % ::getpid() % H); 
    
  SHA_CTX ctx;
  SHA1_Init( &ctx );
  //while ( ( nread = read( fd, buffer, 1024 ) ) > 0 ) {
  SHA1_Update( &ctx, target.c_str(), target.length() );
  //}
  unsigned char bin_sha1_digest[SHA_DIGEST_LENGTH];
  SHA1_Final( bin_sha1_digest, &ctx );
  
  
  target = "";
  bintostring( bin_sha1_digest, SHA_DIGEST_LENGTH, target );//std::string((char *)bin_sha1_digest);
  CREAM_SAFE_LOG(glite::ce::cream_client_api::util::creamApiLogger::instance()->getLogger()->infoStream() 
		   << "certUtil::generateUniqueID() - Generated DelegationID: ["
		   << target << "]"
		   );
}

//___________________________________________________________________________________________
void glite::ce::cream_client_api::certUtil::bintostring( unsigned char* buf, size_t len, std::string& result )
{
  //string result;
  const char alpha[] = "0123456789abcdef";
        
  for ( size_t i=0; i<len; ++i ) {
    result.push_back( alpha[ ( buf[i] >> 4 ) & 0x0f ] );
    result.push_back( alpha[ buf[i] & 0x0f ] );
  }
  
  //return result;
};
